#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <math.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gio/gio.h>

#include "xournal.h"
#include "xo-support.h"
#include "xo-image.h"


// Write pixmap to buffer: returns true on success, false on error.
gboolean pixbuf_to_buffer(GdkPixbuf *pixbuf, gchar **buf, gsize *buflen, const char *type)
{
  GError *error;

  error = NULL;

  if (!gdk_pixbuf_save_to_buffer(pixbuf,
                                 buf,
                                 buflen,
                                 type,
                                 &error,
                                 NULL)) {
    return FALSE;
  }

  return TRUE;
}

// create pixbuf from buffer, or return NULL on failure
GdkPixbuf *pixbuf_from_buffer(const gchar *buf, gsize buflen)
{
  GInputStream *istream;
  GdkPixbuf *pixbuf;
  GError *error;

  error = NULL;

  istream = g_memory_input_stream_new_from_data (buf, buflen, NULL);

  pixbuf = gdk_pixbuf_new_from_stream(istream, NULL, &error); 

  g_input_stream_close(istream, NULL, &error);

  return pixbuf;
}

void create_image_from_pixbuf(GdkPixbuf *pixbuf, struct Item *item, double *pt)
{
  GtkTextBuffer *buffer;
  GnomeCanvasItem *canvas_item;
  GdkColor color;
  GtkWidget *dialog;
  GtkFileFilter *filt_all;
  GtkFileFilter *filt_gdkimage;
  double scale;

  if (item==NULL) {
    item = g_new(struct Item, 1);
    item->type = ITEM_IMAGE;
    item->canvas_item = NULL;
    item->bbox.left = pt[0];
    item->bbox.top = pt[1];
    item->image = pixbuf;
    item->image_scaled = NULL;
    item->image_png = NULL;
    item->image_png_len = 0;

    // Scale at native size, unless that won't fit, 
    // in which case we shrink it down.
    scale = 1 / ui.zoom;
    if( (scale * gdk_pixbuf_get_width(item->image)) > ui.cur_page->width - item->bbox.left ) {
      //set scale so that it does not extend too far to the right
      scale = (ui.cur_page->width - item->bbox.left) / gdk_pixbuf_get_width(item->image);
    }
    if( (scale * gdk_pixbuf_get_height(item->image)) > ui.cur_page->height - item->bbox.top ) {
      //set scale so that it does not extend too far to the bottom
      scale = (ui.cur_page->height - item->bbox.top) / gdk_pixbuf_get_height(item->image);
    }

    item->bbox.right = item->bbox.left + scale * gdk_pixbuf_get_width(item->image);
    item->bbox.bottom = item->bbox.top + scale * gdk_pixbuf_get_height(item->image);
    g_memmove(&(item->brush), ui.cur_brush, sizeof(struct Brush));
    ui.cur_layer->items = g_list_append(ui.cur_layer->items, item);
    ui.cur_layer->nitems++;
  }
  
  item->type = ITEM_IMAGE;
  ui.cur_item = item;
  
  if (item->canvas_item!=NULL) {
    lower_canvas_item_to(ui.cur_layer->group, canvas_item, item->canvas_item);
    gtk_object_destroy(GTK_OBJECT(item->canvas_item));
  }

  make_canvas_item_one(ui.cur_layer->group, item);

  // select image
  reset_selection();
  ui.selection = g_new0(struct Selection, 1);
  ui.selection->type = ITEM_SELECTRECT;
  ui.selection->layer = ui.cur_layer;
  ui.selection->bbox = item->bbox;
  ui.selection->items = g_list_append(ui.selection->items, item);
  ui.selection->canvas_item = gnome_canvas_item_new(ui.cur_layer->group,
      gnome_canvas_rect_get_type(), "width-pixels", 1,
      "outline-color-rgba", 0x000000ff,
      "fill-color-rgba", 0x80808040,
      "x1", ui.selection->bbox.left, "x2", ui.selection->bbox.right, 
      "y1", ui.selection->bbox.top, "y2", ui.selection->bbox.bottom, NULL);
  make_dashed(ui.selection->canvas_item);

  // add undo information
  prepare_new_undo();
  undo->type = ITEM_IMAGE;
  undo->item = ui.cur_item;
  undo->layer = ui.cur_layer;
  
  ui.cur_item = NULL;
  ui.cur_item_type = ITEM_NONE;
}

void insert_image(GdkEvent *event, struct Item *item)
{
  GtkTextBuffer *buffer;
  GnomeCanvasItem *canvas_item;
  GdkColor color;
  GtkWidget *dialog;
  GtkFileFilter *filt_all;
  GtkFileFilter *filt_gdkimage;
  char *filename;
  GdkPixbuf *pixbuf;
  double scale=1;
  double pt[2];
  
  dialog = gtk_file_chooser_dialog_new(_("Insert Image"), GTK_WINDOW (winMain),
     GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
     GTK_STOCK_OPEN, GTK_RESPONSE_OK, NULL);
#ifdef FILE_DIALOG_SIZE_BUGFIX
  gtk_window_set_default_size(GTK_WINDOW(dialog), 500, 400);
#endif
     
  filt_all = gtk_file_filter_new();
  gtk_file_filter_set_name(filt_all, _("All files"));
  gtk_file_filter_add_pattern(filt_all, "*");
  filt_gdkimage = gtk_file_filter_new();
  gtk_file_filter_set_name(filt_gdkimage, _("supported image files"));
  gtk_file_filter_add_pixbuf_formats(filt_gdkimage);
  gtk_file_chooser_add_filter(GTK_FILE_CHOOSER (dialog), filt_gdkimage);
  gtk_file_chooser_add_filter(GTK_FILE_CHOOSER (dialog), filt_all);

  if (ui.default_image != NULL) gtk_file_chooser_set_filename(GTK_FILE_CHOOSER (dialog), ui.default_image);

  if (gtk_dialog_run(GTK_DIALOG(dialog)) != GTK_RESPONSE_OK) {
    gtk_widget_destroy(dialog);
    return;
  }
  filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER (dialog));
  gtk_widget_destroy(dialog);

  if (filename == NULL) {
    /* nothing selected */
    return;
  }

  if (ui.default_image != NULL) g_free(ui.default_image);
  ui.default_image = g_strdup(filename);
  
  pixbuf=gdk_pixbuf_new_from_file(filename, NULL);
  if(pixbuf==NULL){
	  /* open failed */
	  dialog = gtk_message_dialog_new(GTK_WINDOW (winMain), GTK_DIALOG_DESTROY_WITH_PARENT,
	    GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, _("Error opening image '%s'"), filename);
	  gtk_dialog_run(GTK_DIALOG(dialog));
	  gtk_widget_destroy(dialog);
	  g_free(filename);
	  ui.cur_item = NULL;
	  ui.cur_item_type = ITEM_NONE;
	  return;
  }

  ui.cur_item_type = ITEM_IMAGE;

  get_pointer_coords(event, pt);
  set_current_page(pt);  

  create_image_from_pixbuf(pixbuf, item, pt);
}

void update_scaled_image(struct Item *item)
{
  gboolean use_native = is_native_image_usable( item );
  if ( is_native_image_usable( item ) != (item->image_scaled == NULL) ) {
    // we are currently using a scaled image, and should switch to native,
    // or vice versa

    adjust_or_make_canvas_item_image(ui.cur_layer->group, item);
  }
}

void rescale_images(void)
{
  GList *pagelist, *layerlist, *itemlist;
  struct Item *item;
  
  for (pagelist = journal.pages; pagelist!=NULL; pagelist = pagelist->next)
    for (layerlist = ((struct Page *)pagelist->data)->layers; layerlist!=NULL; layerlist = layerlist->next)
      for (itemlist = ((struct Layer *)layerlist->data)->items; itemlist!=NULL; itemlist = itemlist->next) {
        item = (struct Item *)itemlist->data;
        if ( item->type == ITEM_IMAGE ) {
          update_scaled_image( item );
        }
      }
}

gboolean is_native_image_usable(struct Item *item)
{
  double native_width;
  double native_height;
  double required_width;
  double required_height;
  gboolean usable;

  // use the original image at native resolution if it's close enough to what we need
  native_width  = gdk_pixbuf_get_width(item->image) / ui.zoom;
  native_height = gdk_pixbuf_get_height(item->image) / ui.zoom;
  required_width = item->bbox.right - item->bbox.left;
  required_height = item->bbox.bottom - item->bbox.top;
    
  usable = ( fabs( native_width - required_width ) < 1 && 
             fabs( native_height - required_height ) < 1 );

  return usable;
}
